#!/bin/bash



# 
# NAME
#     gitosis-add-repo.sh - Add new repository to gitosis
# 
# 
# SYNOPSIS
#     gitosis-add-repo.sh [-v] [-l] [-t TEMP_DIR] [-u GITOSIS_USER] [-h
#     GITOSIS_HOST] [-k SSH_PUB_KEY] [PATH_GIT_PROJECT]
# 
# 
# DESCRIPTION
#     This program is used to add a new repository to the gitosis program
#     (http://eagain.net/gitweb/?p=gitosis.git). If no path to the git project
#     is provided, the program takes the current directory.
# 
# 
# OPTIONS
#     -v      Enables verbose output
# 
#     -l      Enables logging into syslog via logger
# 
#     -t TEMP_DIR
#             Use TEMP_DIR instead of the one that is configured in $TEMP_DIR
#             (usually ~/.tmp_gitosis)
# 
#     -u GITOSIS_USER
#             Use GITOSIS_USER instead of the one that is configured in
#             $GITOSIS_USER (usually gitosis)
# 
#     -h GITOSIS_HOST
#             Use GITOSIS_HOST instead of the one that is configured in
#             $GITOSIS_HOST (usually gitosis)
# 
#     -k SSH_PUB_KEY
#             File containing the public key to be used for SSH access to
#             gitosis
# 
# 
# OPERATION
#     The following steps will be performed
# 
#     *   Read user config file. This is done by simply sourcing in
#         ~/.gitosis-scripts
# 
#     *   Evaluate command line options. Since this happens after reading the
#         user config file, settings provided on the command line will always
#         override values from the config file
# 
#     *   Check whether required programs are available (in particular git and
#         logger)
# 
#     *   Check whether the machine that hosts gitosis is reachable
# 
#     *   Determine repo path and name. If no path was specified on the
#         command line this will be the current directory
# 
#     *   Check whether a git repo exists in the path
# 
#     *   Extract user from SSH public key file
# 
#     *   Create temp dir (usually ~/.gitosis_tmp). If this exists already, it
#         will be emptied
# 
#     *   Clone (check out) gitosis-admin into temp dir
# 
#     *   Check whether project of the same name already exists in gitosis
# 
#     *   Add project to gitosis and push back (check in) gitosis-admin
# 
#     *   Add file .gitosis_host to project that contains the hostname of
#         gitosis
# 
#     *   Push the project itself (so far only the updated gitosis-admin had
#         made it back to gitosis) to gitosis
# 
# 
# AUTHOR
#     Christoph Jahn <http://www.christoph-jahn.com>
# 
# 
# COPYRIGHT AND LICENSE
#     Copyright 2009 by Christoph Jahn <http://www.christoph-jahn.com>
# 
#     This program is free software: you can redistribute it and/or modify it
#     under the terms of the GNU General Public License as published by the
#     Free Software Foundation, either version 3 of the License, or (at your
#     option) any later version.
# 
#     This program is distributed in the hope that it will be useful, but
#     WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
#     Public License for more details.
# 
#     You should have received a copy of the GNU General Public License along
#     with this program. If not, see <http://www.gnu.org/licenses/>.
# 
# 
# SEE ALSO
#     This program is basically the coded equivalent of the steps that are
#     described in
#     http://scie.nti.st/2007/11/14/hosting-git-repositories-the-easy-and-secu
#     re-way
# 





# ==================================================
#
# No configuration should be needed below this point
#
# If you want to change defaults, please do so in 
# the config file, which is defined in $CFG_FILE
#
# ==================================================



# Temp directory used to check out the gitosis-admin.git project. 
# Will automatically be created if it does not exists.
# If it exists, all files in it will automatically be deleted
TEMP_DIR=~/.tmp_gitosis


# Host where gitosis is running on
GITOSIS_HOST=gitosis

# Algorithm used for computing the keys. Should be "rsa" or "dsa"
SSH_KEY_ALGORITHM=rsa

# File that contains the SSH public key used for accessing gitosis
SSH_PUB_KEY=~/.ssh/id_$SSH_KEY_ALGORITHM.pub


# This is the user for accessing gitosis. This is a technical user
# and should not be changed to any real user. Other possible values
# (depending on your setup of gitosis) might be git, vcs etc.
# On Debian Lenny this should be set to gitosis.
GITOSIS_USER=gitosis


# By default output to syslog via 'logger' is disabled
# Use '-l' switch to enable it. This sets $SYSLOG_OUTPUT=1
SYSLOG_OUTPUT=0
SYSLOG_FACILITY=user


# Determines whether debug information is sent to STDOUT
#   disabled: VERBOSE=0
#   enabled : VERBOSE=1
# Default is disabled. Can be enabled with the '-v' switch
# on the command line
VERBOSE=1


PRG=`basename $0`

# This is name of the git project that contains the
# runtime configuration of gitosis (gitosis is 
# maintaining itself in a "recursive" manner)
GITOSIS_ADMIN=gitosis-admin


# The only real reason that the following commands (CMD_*) are defined
# as variables is the DRY (don't repeat yourself) principle.  This is 
# necessary for checking whether the programs are in the path. They are
# not expected to change any time soon. 
CMD_GIT=git
CMD_SYSLOG=logger

# This array is only used to check whether the specified
# programs actually exist on the system and are in the path.
# This is done by issuing the 'which' command and checking
# its return code.
CMD_CHECK[0]=$CMD_GIT
CMD_CHECK[1]=$CMD_SYSLOG



# Return codes used to indicate various errors detected.

RC_INVALID_CMDLINE_OPTION=14
RC_INVALID_PATH_REPO=5
RC_GITOSIS_HOST_UNREACHABLE=7
RC_SSH_PUB_KEY_NOT_FOUND=8
RC_SSH_PUB_KEY_NOT_READABLE=9
RC_NO_GIT_REPO_FOUND=10
RC_REPO_IN_GITOSIS=13






# Handle output of messages.
#
# Usage:   output { err | info | debug } MESSAGE
# 
# Level:  err     Send to STDERR
#         info    Send to STDOUT
#         debug   Send to STDOUT if $VERBOSE=1
#
# Syslog: If $SYSLOG_OUTPUT=1 send to syslog via 'logger'
#         (facility = $SYSLOG_FACILITY)
#
function output
{
    PRIO=$1
    MSG="$2"
    if [ "$PRIO" = "info" ]; then
	echo "$PRG: $MSG"
    elif [ "$PRIO" = "err" ]; then
	echo "$PRG: $MSG" >&2
    elif [ "$PRIO" = "debug" -a "$VERBOSE" -eq 1 ]; then
	echo "$PRG: $MSG"
    fi

    # Send to syslog if this is enabled
    if [ $SYSLOG_OUTPUT -eq 1 ]; then
	$CMD_SYSLOG -p $SYSLOG_FACILITY.$PRIO -t "$PRG" "$MSG"
    fi
}




# Output message with return code and exit
#
# Usage:   abort RC
#
function abort
{
    RC="$1"
    if [ "$VERBOSE" -eq 1 ]; then
	output err "Aborting ... (RC=$RC)"
    fi
    exit $RC
}







# -------------------------------
# Read config file
# -------------------------------

RC_CANNOT_READ_CFG_FILE=1
RC_CANNOT_CREATE_CFG_FILE=2
RC_CANNOT_SET_PERMISSIONS_CFG_FILE=3
RC_CFG_FILE_NOT_FOUND=4
RC_CANNOT_POPULATE_CFG_FILE=16
RC_NEW_CFG_FILE_NEEDS_CHECK=17




function read_user_config
{

    CFG_FILE="$1"
    CFG_TEMPLATE="$2"
    CFG_CREATE_IF_EMPTY="$3"

    PERMISSIONS=600


    if [ -f $CFG_FILE ]; then
	if [ ! -r $CFG_FILE ]; then
	    output err "Cannot read config file ($CFG_FILE)"
	    abort $RC_CANNOT_READ_CFG_FILE
	else
	    . $CFG_FILE
	    output debug "Started"
	    output debug "Sourcing in config file ($CFG_FILE)"
	fi
    else

	# Should user config file be created with template values?
	if [ "$CFG_CREATE_IF_EMPTY" -eq 1 ]; then

            # Create empty config file (content comes in later, once permissions
	    # were set appropriately; this is a security feature)
	    output info "No config file ($CFG_FILE) found, creating one with default values"
	    touch $CFG_FILE
	    RET=$?
	    if [ "$RET" -ne 0 ]; then
		output err "Error while trying to create config file ($CFG_FILE)"
		abort $RC_CANNOT_CREATE_CFG_FILE
	    else
		output err "Successfully created empty config file ($CFG_FILE)"
	    fi

            # Change access rights for config file to make it private
	    chmod $PERMISSIONS $CFG_FILE
	    RET=$?
	    if [ "$RET" -ne 0 ]; then
		output err "Error while trying to change permissions of config file ($CFG_FILE) to $PERMISSIONS"
		abort $RC_CANNOT_SET_PERMISSIONS_CFG_FILE
	    else
		output debug "Changed permissions of config file ($CFG_FILE) to $PERMISSIONS"
	    fi

            # Fill config file with default values
	    cat <<EOF  >> $CFG_FILE
# Temp directory used to check out the gitosis-admin.git project. 
# Will automatically be created if it does not exists.
# If it exists, all files in it will automatically be deleted
TEMP_DIR=~/.tmp_gitosis


# Host where gitosis is running on
GITOSIS_HOST=gitosis

# Algorithm used for computing the keys. Should be "rsa" or "dsa"
SSH_KEY_ALGORITHM=rsa

# File that contains the SSH public key used for accessing gitosis
SSH_PUB_KEY=~/.ssh/id_$SSH_KEY_ALGORITHM.pub


# This is the user for accessing gitosis. This is a technical user
# and should not be changed to any real user. Other possible values
# (depending on your setup of gitosis) might be git, vcs etc.
# On Debian Lenny this should be set to gitosis.
GITOSIS_USER=gitosis


# By default output to syslog via 'logger' is disabled
# Use '-l' switch to enable it. This sets $SYSLOG_OUTPUT=1
SYSLOG_OUTPUT=0
SYSLOG_FACILITY=user


# Determines whether debug information is sent to STDOUT
#   disabled: VERBOSE=0
#   enabled : VERBOSE=1
# Default is disabled. Can be enabled with the '-v' switch
# on the command line
VERBOSE=1

EOF
	    RET=$?
	    if [ "$RET" -ne 0 ]; then
		output err "Error while trying to populate config file ($CFG_FILE)"
		abort $RC_CANNOT_POPULATE_CFG_FILE
	    else
		output err "Successfully populated config file ($CFG_FILE)"
	    fi
	    output info "New user config file ($CFG_FILE) was created. Please update it according to your needs (especially check the value of GITOSIS_HOST). This program will now end so that you can perform this check"
	    abort $RC_NEW_CFG_FILE_NEEDS_CHECK


	# User config does not exist but shall not be created either, this is an error
	else
	    output err "Config file ($CFG_FILE) not found"
	    abort $RC_CFG_FILE_NOT_FOUND
	fi
    fi
}

read_user_config ~/.gitosis-scripts gitosis-scripts_config.sh 1


# -------------------------------
# Evaluate command line
# -------------------------------

while getopts 'vltu:h:k:' OPTION
do
    case $OPTION in
	v)	VERBOSE=1
	    output debug "Started"
	    ;;
	l)	SYSLOG_OUTPUT=1
	    ;;
	t)	TEMP_DIR="$OPTARG"
	    ;;
	u)	GITOSIS_USER="$OPTARG"
	    ;;
	h)  GITOSIS_HOST="$OPTARG"
	    output debug "Using gitosis on host '$GITOSIS_HOST'"
	    ;;
	k)	SSH_PUB_KEY="$OPTARG"
	    ;;
	?)	printf 'Usage : gitosis-add-repo.sh [B<-v>] [B<-l>] [B<-t> I<TEMP_DIR>] [B<-u> I<GITOSIS_USER>] [B<-h> I<GITOSIS_HOST>] [B<-k> I<SSH_PUB_KEY>] [PATH_GIT_PROJECT]
 \n' $(basename $0) >&2
	    abort $RC_INVALID_CMDLINE_OPTION
	    ;;
    esac
done
shift $(($OPTIND - 1))




RC_PROGRAM_MISSING=6


# -------------------------------
# Check for required programs
# -------------------------------

output debug "Checking whether programs are executable and in the path"
index=0
for i in ${CMD_CHECK[@]}
do
    CMD_RET=`which ${CMD_CHECK[$index]} > /dev/null;echo $?`
    if [ "$CMD_RET" = "0" ] ; then
	output debug "- Found program '${CMD_CHECK[$index]}'"
    else
	output err "Program '${CMD_CHECK[$index]}' cannot be found"
	abort $RC_PROGRAM_MISSING
    fi
    (( index = index + 1 ))
done
output debug "All required programs were found"



# -------------------------------
# gitosis host reachable?
# -------------------------------

output debug "Checking connectivity to gitosis host '$GITOSIS_HOST'"
GITOSIS_HOST_STATUS=`ping -c 3 $GITOSIS_HOST > /dev/null; echo $?`
if [ "$GITOSIS_HOST_STATUS" -ne 0 ]; then
    output err "Cannot reach gitosis host '$GITOSIS_HOST'"
    abort $RC_GITOSIS_HOST_UNREACHABLE
else
    output debug "Cconnectivity to gitosis host '$GITOSIS_HOST' is ok"
fi




# -------------------------------
# Determine path of git repo and check
# -------------------------------

# Determine path and check existence
PATH_GIT_PROJECT="$1"
if [ "$PATH_GIT_PROJECT" == "" ]; then
    PATH_GIT_PROJECT=`pwd`
    output debug "No path provided for git repo, using current directory"
else
    if [ ! -d "$PATH_GIT_PROJECT" ]; then
	output err "Invalid path ($PATH_GIT_PROJECT) for git repo"
	abort $RC_INVALID_PATH_REPO
    else
	output debug "Path of git repo ($PATH_GIT_PROJECT) is valid"
    fi
fi

# Does path contain valid git repo?
REPO_STATUS=`cd $PATH_GIT_PROJECT; $CMD_GIT status > /dev/null 2>&1 ; echo $?`
if [  "$REPO_STATUS" -eq 128 ] ; then
    output err "No valid git repo found in '$PATH_GIT_PROJECT'"
    abort $RC_NO_GIT_REPO_FOUND
else
    output debug "Valid git repo found in '$PATH_GIT_PROJECT'"
fi
GIT_PROJECT=`basename $PATH_GIT_PROJECT`
output debug "Repo name for gitosis set to '$GIT_PROJECT'"



# -------------------------------
# Get user from SSH public key
# -------------------------------

if [ ! -f "$SSH_PUB_KEY" ]; then
    output err "File with SSH public key ($SSH_PUB_KEY) does not exist"
    abort $RC_SSH_PUB_KEY_NOT_FOUND
elif [ ! -r "$SSH_PUB_KEY" ]; then
    output err "File with SSH public key ($SSH_PUB_KEY) cannot be read"
    abort $RC_SSH_PUB_KEY_NOT_READABLE
fi
REPO_USER=`gawk '{print $3}' "$SSH_PUB_KEY"`
output debug "Using file '$SSH_PUB_KEY' for SSH public key; user = '$REPO_USER'"



RC_TEMP_DIR_UNWRITABLE=11
RC_TEMP_DIR_NOT_EMPTY=12
RC_TEMP_DIR_NOT_CREATED=15


# -------------------------------
# Create temp dir
# -------------------------------

function create_temp_dir
{
    TEMP_DIR="$1"
    CURRENT_USER=`whoami`

    # Is temp directory writable?
    if [  -d "$TEMP_DIR"   -a  ! -w "$TEMP_DIR"  ] ; then
	output err "Temp dir '$TEMP_DIR' already exists but is not writable for user '$CURRENT_USER'"
	abort $RC_TEMP_DIR_UNWRITABLE
    fi

    # Clean up 
    if [ -d "$TEMP_DIR" ]; then
	output info "Temp dir '$TEMP_DIR' already exists, cleaning up"
	rm -rf "$TEMP_DIR"/*
    else
	output info "Creating temp dir '$TEMP_DIR'"
	mkdir "$TEMP_DIR"
	RET=$?
	if [ "$RET" -ne 0 ]; then
	    output err "New temp dir '$TEMP_DIR' could not be created"
	    abort $RC_TEMP_DIR_NOT_CREATED
	fi
    fi

    # Any files left in temp dir?
    FILES_IN_TEMP=`ls -A $TEMP_DIR | wc -l`
    if [ $FILES_IN_TEMP -gt 0 ]; then
	output err "Temp dir '$TEMP_DIR' is not empty"
	abort $RC_TEMP_DIR_NOT_EMPTY
    fi
}
create_temp_dir "$TEMP_DIR"
cd "$TEMP_DIR"


# -------------------------------
# Add git repo to gitosis.conf
# -------------------------------

# Get gitosis-admin
output debug "Cloning gitosis-admin from host '$GITOSIS_HOST'"
$CMD_GIT clone $GITOSIS_USER@$GITOSIS_HOST:$GITOSIS_ADMIN.git
cd $TEMP_DIR/$GITOSIS_ADMIN;


# Check whether project has already been added
output debug "Checking whether repo '$GIT_PROJECT' already exists"
REPO_STATUS=`grep "writable[ \t]=[ \t]$GIT_PROJECT" gitosis.conf>/dev/null; echo $? `

if [ "$REPO_STATUS" -eq 0 ]; then
    output err "Repo '$GIT_PROJECT' already exists in gitosis"
    abort $RC_REPO_IN_GITOSIS
fi


# Add repo to gitosis
output debug "Adding repo '$GIT_PROJECT' to gitosis.conf"

echo " " >> gitosis.conf
echo "[group $GIT_PROJECT]" >> gitosis.conf
echo "writable = $GIT_PROJECT" >> gitosis.conf
echo "members = $REPO_USER" >> gitosis.conf

# Commit and push back to gitosis
MSG_DONE="Added repo '$GIT_PROJECT' with user '$REPO_USER' to gitosis"
$CMD_GIT commit -a -m "$MSG_DONE"
$CMD_GIT push
output info "$MSG_DONE"


# -------------------------------
# Push git repo to gitosis
# -------------------------------

#
OLD_DIR=`pwd`
output debug "Configuring git repo for being pushed to gitosis"
cd $PATH_GIT_PROJECT
$CMD_GIT remote add origin $GITOSIS_USER@$GITOSIS_HOST:$GIT_PROJECT.git

# do some work, git add and commit files
echo "$GITOSIS_HOST" > .gitosis_host
$CMD_GIT add .gitosis_host
$CMD_GIT commit -a -m "Added file containing gitosis host" --no-verify

output debug "Pushing git repo to gitosis"
$CMD_GIT push origin master:refs/heads/master

cd $OLD_DIR

output debug "Finished"
exit 0

