#!/bin/bash



# 
# NAME
#     gitosis-init.sh - Initial setup of gitosis
# 
# 
# SYNOPSIS
#     gitosis-init.sh [-v] [-l] [-t TEMP_DIR] [-u GITOSIS_USER] SSH_PUB_KEY
# 
# 
# DESCRIPTION
#     This program is used to perform the initial setup of the gitosis program
#     (http://eagain.net/gitweb/?p=gitosis.git). It needs a file containing
#     the public key for the user you want to use as initial admin user for
#     gitosis. Only this user will initially be able to work with gitosis. If
#     you want to add other users later, use this initial user to do so.
# 
# 
# OPTIONS
#     -v      Enables verbose output
# 
#     -l      Enables logging into syslog via logger
# 
#     -t TEMP_DIR
#             Use TEMP_DIR instead of the one that is configured in $TEMP_DIR
#             (usually ~/.tmp_gitosis)
# 
#     -u GITOSIS_USER
#             Use GITOSIS_USER instead of the one that is configured in
#             $GITOSIS_USER (usually gitosis)
# 
# 
# OPERATION
#     The following steps will be performed
# 
#     *   Read user config file. This is done by simply sourcing in
#         ~/.gitosis-scripts.
# 
#     *   Evaluate command line options. Since this happens after reading the
#         user config file, settings provided on the command line will always
#         override values from the config file.
# 
#     *   Check whether required programs are available (in particular git and
#         logger).
# 
#     *   Check whether required user exists.
# 
#     *   Check whether gitosis has already been initialized.
# 
#     *   Check whether current user is either root (recommended) or
#         GITOSIS_USER.
# 
#     *   Check whether SSH public key file that was provided on the command
#         line exists and can be read. This is the public key of the user that
#         you will primarily be working with.
# 
#         ATTENTION: This is NOT the public key of GITOSIS_USER.
# 
#     *   Create temp dir.
# 
#     *   If SSH public key of GITOSIS_USER does not exist, create it. This is
#         only needed for the initial setup activity of this script. You will
#         not work with this key later.
# 
#     *   Initiailze gitosis with public key of GITOSIS_USER (the public key
#         you provided via command line will be added later).
# 
#     *   Clone (check out) gitosis-admin and add public key from command line
#         to repo. Also add user specified in this key as admin to
#         gitosis-admin.
# 
#     *   Push (check in) gitosis-admin with the above changes. You can now
#         start working on it remotely with the user whose SSH public key was
#         specified on the command line.
# 
# 
# FILES
#     The following files are specific to gitosis-init.sh
# 
#     * ~/.gitosis-scripts User configuration
# 
#     * ~/.gitosis_tmp Temp directory default
# 
#     In addition the following binaries need to be installed on the system
#     and be located in the PATH
# 
#     * logger
# 
#     * gitosis-init
# 
#     * git
# 
#     * ssh-keygen
# 
# 
# AUTHOR
#     Christoph Jahn <http://www.christoph-jahn.com>
# 
# 
# COPYRIGHT AND LICENSE
#     Copyright 2009 by Christoph Jahn <http://www.christoph-jahn.com>
# 
#     This program is free software: you can redistribute it and/or modify it
#     under the terms of the GNU General Public License as published by the
#     Free Software Foundation, either version 3 of the License, or (at your
#     option) any later version.
# 
#     This program is distributed in the hope that it will be useful, but
#     WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
#     Public License for more details.
# 
#     You should have received a copy of the GNU General Public License along
#     with this program. If not, see <http://www.gnu.org/licenses/>.
# 
# 
# SEE ALSO
#     This program is basically the coded equivalent of the steps that are
#     described in
#     http://scie.nti.st/2007/11/14/hosting-git-repositories-the-easy-and-secu
#     re-way
# 




# ==================================================
#
# No configuration should be needed below this point
#
# If you want to change defaults, please do so in 
# the config file, which is defined in $CFG_FILE
#
# ==================================================



# Temp directory used to check out the gitosis-admin.git project. 
# Will automatically be created if it does not exists.
# If it exists, all files in it will automatically be deleted
TEMP_DIR=~/.tmp_gitosis


# Host where gitosis is running on
GITOSIS_HOST=gitosis

# Algorithm used for computing the keys. Should be "rsa" or "dsa"
SSH_KEY_ALGORITHM=rsa

# File that contains the SSH public key used for accessing gitosis
SSH_PUB_KEY=~/.ssh/id_$SSH_KEY_ALGORITHM.pub


# This is the user for accessing gitosis. This is a technical user
# and should not be changed to any real user. Other possible values
# (depending on your setup of gitosis) might be git, vcs etc.
# On Debian Lenny this should be set to gitosis.
GITOSIS_USER=gitosis


# By default output to syslog via 'logger' is disabled
# Use '-l' switch to enable it. This sets $SYSLOG_OUTPUT=1
SYSLOG_OUTPUT=0
SYSLOG_FACILITY=user


# Determines whether debug information is sent to STDOUT
#   disabled: VERBOSE=0
#   enabled : VERBOSE=1
# Default is disabled. Can be enabled with the '-v' switch
# on the command line
VERBOSE=1





PRG=`basename $0`

GITOSIS_ADMIN=gitosis-admin

CMD_GIT=git
CMD_GITOSIS_INIT=gitosis-init
CMD_SSH_KEYGEN=ssh-keygen
CMD_SYSLOG=logger

# This array is only used to check whether the specified
# programs actually exist on the system and are in the path.
# This is done by issuing the 'which' command and checking
# its return code.
CMD_CHECK[0]=$CMD_GIT
CMD_CHECK[1]=$CMD_GITOSIS_INIT
CMD_CHECK[2]=$CMD_SSH_KEYGEN
CMD_CHECK[3]=$CMD_SYSLOG





# Return codes used to indicate various errors detected.
RC_INVALID_CMDLINE_OPTION=25
RC_USER_MISSING=24
RC_NOT_FIRST_RUN=26
RC_WRONG_USER=27
RC_PROGRAM_MISSING=28
RC_CMD_PARAM_MISSING=21
RC_SSH_KEY_UNREADABLE=5
RC_SSH_KEY_MISSING=22
RC_TEMP_DIR_UNWRITABLE=23
RC_TEMP_DIR_NOT_EMPTY=8
RC_ERR_SSH_KEY_CREATION=9
RC_USER_ALREADY_GITOSIS_ADMIN=10







# Handle output of messages.
#
# Usage:   output { err | info | debug } MESSAGE
# 
# Level:  err     Send to STDERR
#         info    Send to STDOUT
#         debug   Send to STDOUT if $VERBOSE=1
#
# Syslog: If $SYSLOG_OUTPUT=1 send to syslog via 'logger'
#         (facility = $SYSLOG_FACILITY)
#
function output
{
    PRIO=$1
    MSG="$2"
    if [ "$PRIO" = "info" ]; then
	echo "$PRG: $MSG"
    elif [ "$PRIO" = "err" ]; then
	echo "$PRG: $MSG" >&2
    elif [ "$PRIO" = "debug" -a "$VERBOSE" -eq 1 ]; then
	echo "$PRG: $MSG"
    fi

    # Send to syslog if this is enabled
    if [ $SYSLOG_OUTPUT -eq 1 ]; then
	$CMD_SYSLOG -p $SYSLOG_FACILITY.$PRIO -t "$PRG" "$MSG"
    fi
}




# Output message with return code and exit
#
# Usage:   abort RC
#
function abort
{
    RC="$1"
    if [ "$VERBOSE" -eq 1 ]; then
	output err "Aborting ... (RC=$RC)"
    fi
    exit $RC
}







# -------------------------------
# Read config file
# -------------------------------

RC_CANNOT_READ_CFG_FILE=1
RC_CANNOT_CREATE_CFG_FILE=2
RC_CANNOT_SET_PERMISSIONS_CFG_FILE=3
RC_CFG_FILE_NOT_FOUND=4
RC_CANNOT_POPULATE_CFG_FILE=16
RC_NEW_CFG_FILE_NEEDS_CHECK=17




function read_user_config
{

    CFG_FILE="$1"
    CFG_TEMPLATE="$2"
    CFG_CREATE_IF_EMPTY="$3"

    PERMISSIONS=600


    if [ -f $CFG_FILE ]; then
	if [ ! -r $CFG_FILE ]; then
	    output err "Cannot read config file ($CFG_FILE)"
	    abort $RC_CANNOT_READ_CFG_FILE
	else
	    . $CFG_FILE
	    output debug "Started"
	    output debug "Sourcing in config file ($CFG_FILE)"
	fi
    else

	# Should user config file be created with template values?
	if [ "$CFG_CREATE_IF_EMPTY" -eq 1 ]; then

            # Create empty config file (content comes in later, once permissions
	    # were set appropriately; this is a security feature)
	    output info "No config file ($CFG_FILE) found, creating one with default values"
	    touch $CFG_FILE
	    RET=$?
	    if [ "$RET" -ne 0 ]; then
		output err "Error while trying to create config file ($CFG_FILE)"
		abort $RC_CANNOT_CREATE_CFG_FILE
	    else
		output err "Successfully created empty config file ($CFG_FILE)"
	    fi

            # Change access rights for config file to make it private
	    chmod $PERMISSIONS $CFG_FILE
	    RET=$?
	    if [ "$RET" -ne 0 ]; then
		output err "Error while trying to change permissions of config file ($CFG_FILE) to $PERMISSIONS"
		abort $RC_CANNOT_SET_PERMISSIONS_CFG_FILE
	    else
		output debug "Changed permissions of config file ($CFG_FILE) to $PERMISSIONS"
	    fi

            # Fill config file with default values
	    cat <<EOF  >> $CFG_FILE
# Temp directory used to check out the gitosis-admin.git project. 
# Will automatically be created if it does not exists.
# If it exists, all files in it will automatically be deleted
TEMP_DIR=~/.tmp_gitosis


# Host where gitosis is running on
GITOSIS_HOST=gitosis

# Algorithm used for computing the keys. Should be "rsa" or "dsa"
SSH_KEY_ALGORITHM=rsa

# File that contains the SSH public key used for accessing gitosis
SSH_PUB_KEY=~/.ssh/id_$SSH_KEY_ALGORITHM.pub


# This is the user for accessing gitosis. This is a technical user
# and should not be changed to any real user. Other possible values
# (depending on your setup of gitosis) might be git, vcs etc.
# On Debian Lenny this should be set to gitosis.
GITOSIS_USER=gitosis


# By default output to syslog via 'logger' is disabled
# Use '-l' switch to enable it. This sets $SYSLOG_OUTPUT=1
SYSLOG_OUTPUT=0
SYSLOG_FACILITY=user


# Determines whether debug information is sent to STDOUT
#   disabled: VERBOSE=0
#   enabled : VERBOSE=1
# Default is disabled. Can be enabled with the '-v' switch
# on the command line
VERBOSE=1

EOF
	    RET=$?
	    if [ "$RET" -ne 0 ]; then
		output err "Error while trying to populate config file ($CFG_FILE)"
		abort $RC_CANNOT_POPULATE_CFG_FILE
	    else
		output err "Successfully populated config file ($CFG_FILE)"
	    fi
	    output info "New user config file ($CFG_FILE) was created. Please update it according to your needs (especially check the value of GITOSIS_HOST). This program will now end so that you can perform this check"
	    abort $RC_NEW_CFG_FILE_NEEDS_CHECK


	# User config does not exist but shall not be created either, this is an error
	else
	    output err "Config file ($CFG_FILE) not found"
	    abort $RC_CFG_FILE_NOT_FOUND
	fi
    fi
}

read_user_config ~/.gitosis-scripts gitosis-scripts_config.sh 1

PWD=`pwd`

function run_cmd
{
    CMD="$*"
    USER=`whoami`
    if [ "$USER" != "root" ]; then
#	output info "Running directly as user $USER"
#	su -l "$GITOSIS_USER" -c "$CMD"
	exec $CMD
	RET=$?
    else
	su -l "$GITOSIS_USER" -c "$CMD"
	RET=$?
    fi
    if [ "$RET" != "0" ] ;then
	output err "Error (RC=$RET) running command '$CMD'"
	exit 1
    fi
}





# -------------------------------
# Evaluate command line
# -------------------------------

while getopts 'vltu:' OPTION
do
    case $OPTION in
	l)	SYSLOG_OUTPUT=1
	    ;;
	v)	VERBOSE=1
	    ;;
	t)	TEMP_DIR="$OPTARG"
	    ;;
	u)	GITOSIS_USER="$OPTARG"
	    ;;
	?)	printf 'Usage : gitosis-init.sh [B<-v>] [B<-l>] [B<-t> I<TEMP_DIR>] [B<-u> I<GITOSIS_USER>] SSH_PUB_KEY
 \n' $(basename $0) >&2
	    exit $RC_INVALID_CMDLINE_OPTION
	    ;;
    esac
done
shift $(($OPTIND - 1))



output info "Started setting up GITOSIS for initial use"





RC_PROGRAM_MISSING=6


# -------------------------------
# Check for required programs
# -------------------------------

output debug "Checking whether programs are executable and in the path"
index=0
for i in ${CMD_CHECK[@]}
do
    CMD_RET=`which ${CMD_CHECK[$index]} > /dev/null;echo $?`
    if [ "$CMD_RET" = "0" ] ; then
	output debug "- Found program '${CMD_CHECK[$index]}'"
    else
	output err "Program '${CMD_CHECK[$index]}' cannot be found"
	abort $RC_PROGRAM_MISSING
    fi
    (( index = index + 1 ))
done
output debug "All required programs were found"





# -------------------------------
# Check whether required user exists
# -------------------------------

USER_NOT_FOUND=`grep $GITOSIS_USER /etc/passwd > /dev/null; echo $?`
if [  "$USER_NOT_FOUND" -eq 1 ];then
    output err "User $GITOSIS_USER does not exist on this system."
    abort $RC_USER_MISSING
fi




# -------------------------------
# Check for previous run
# -------------------------------

GITOSIS_USER_HOMEDIR=`grep $GITOSIS_USER /etc/passwd | awk -F : '{ print $6 }'`
output debug "Home dir of user '$GITOSIS_USER' = $GITOSIS_USER_HOMEDIR"
if [ -d "$GITOSIS_USER_HOMEDIR/gitosis" -o -d "$GITOSIS_USER_HOMEDIR/repositories" ];then
    output err "Found $GITOSIS_USER_HOMEDIR/{gitosis|repositories}, so it seems that gitosis has already been initialized."
    abort $RC_NOT_FIRST_RUN
fi




# -------------------------------
# Check current user
# -------------------------------


# Program needs to run with user root or $GITOSIS_USER
CURRENT_USER=`whoami`
if [  "$CURRENT_USER" == "root"  -o  "$CURRENT_USER" == "$GITOSIS_USER"  ] ; then
    output debug "Current user '$CURRENT_USER' is valid"
else
    output err "Current user '$CURRENT_USER' is not valid. User must be either 'root' or '$GITOSIS_USER'."
    abort $RC_WRONG_USER
fi






# -------------------------------
# Get SSH pub key file
# -------------------------------



# Get filename for pub SSH key of user that will be used to
# interact with gitosis
PUB_SSH_KEY="$1"
if [ "$PUB_SSH_KEY" == "" ]; then
    output err "Missing parameter: File name for public key"
    abort $RC_CMD_PARAM_MISSING
fi

if [  -f "$PUB_SSH_KEY" ] ; then
    if [ ! -r "$PUB_SSH_KEY" ] ; then
	output err "File with SSH public key ($PUB_SSH_KEY) exists, but cannot be read"
	abort $RC_SSH_KEY_UNREADABLE
    fi
    REPO_USER=`gawk '{print $3}' "$PUB_SSH_KEY"`
    output debug "Using file '$PUB_SSH_KEY' for SSH public key"
else
    output err "File with SSH public key ($PUB_SSH_KEY) does not exist or cannot be read"
    abort $RC_SSH_KEY_MISSING
fi



RC_TEMP_DIR_UNWRITABLE=11
RC_TEMP_DIR_NOT_EMPTY=12
RC_TEMP_DIR_NOT_CREATED=15


# -------------------------------
# Create temp dir
# -------------------------------

function create_temp_dir
{
    TEMP_DIR="$1"
    CURRENT_USER=`whoami`

    # Is temp directory writable?
    if [  -d "$TEMP_DIR"   -a  ! -w "$TEMP_DIR"  ] ; then
	output err "Temp dir '$TEMP_DIR' already exists but is not writable for user '$CURRENT_USER'"
	abort $RC_TEMP_DIR_UNWRITABLE
    fi

    # Clean up 
    if [ -d "$TEMP_DIR" ]; then
	output info "Temp dir '$TEMP_DIR' already exists, cleaning up"
	rm -rf "$TEMP_DIR"/*
    else
	output info "Creating temp dir '$TEMP_DIR'"
	mkdir "$TEMP_DIR"
	RET=$?
	if [ "$RET" -ne 0 ]; then
	    output err "New temp dir '$TEMP_DIR' could not be created"
	    abort $RC_TEMP_DIR_NOT_CREATED
	fi
    fi

    # Any files left in temp dir?
    FILES_IN_TEMP=`ls -A $TEMP_DIR | wc -l`
    if [ $FILES_IN_TEMP -gt 0 ]; then
	output err "Temp dir '$TEMP_DIR' is not empty"
	abort $RC_TEMP_DIR_NOT_EMPTY
    fi
}
create_temp_dir "$TEMP_DIR"
cd "$TEMP_DIR"



chown $GITOSIS_USER "$TEMP_DIR"




# -------------------------------
# Create SSH keys
# -------------------------------

# Create SSH keys for gitosis user if necessary
GITOSIS_SSH_KEY=$GITOSIS_USER_HOMEDIR/.ssh/id_$SSH_KEY_ALGORITHM

output debug "Checking if SSH keys for inital admin user ($GITOSIS_USER) already exist (file=$GITOSIS_SSH_KEY.pub)"
if [ -f "$GITOSIS_SSH_KEY.pub" ]; then
    output debug "SSH keys for inital admin user ($GITOSIS_USER) already exist, will be re-used"
else
    output debug "Creating SSH keys for inital admin user ($GITOSIS_USER)"
    run_cmd "$CMD_SSH_KEYGEN -t $SSH_KEY_ALGORITHM -f $GITOSIS_SSH_KEY"
    if [ ! -f "$GITOSIS_SSH_KEY.pub" -o ! -f "$GITOSIS_SSH_KEY.pub" ]; then
	output err "Failure while creating SSH key files."
	abort $RC_ERR_SSH_KEY_CREATION
    fi
fi




# -------------------------------
# Init gitosis
# -------------------------------

# Do the actual initialization
output debug "Initializing Git repository"
run_cmd "$CMD_GITOSIS_INIT < $GITOSIS_SSH_KEY.pub"
output debug "Finished setting up GITOSIS for initial use"



# -------------------------------
# Add user as admin
# -------------------------------

output debug "Adding user '$REPO_USER' as admin (this is the user you will be working with)"
run_cmd "cd $TEMP_DIR; $CMD_GIT clone $GITOSIS_USER@$GITOSIS_HOST:$GITOSIS_ADMIN.git"
cd $TEMP_DIR/$GITOSIS_ADMIN;

# Check whether user has already been added
if [ -f $TEMP_DIR/$GITOSIS_ADMIN/keydir/$REPO_USER.pub ]; then
    output err "File with user key '$REPO_USER' is already in gitosis key directory. Most likely the repo has already been initialized."
    abort $RC_USER_ALREADY_GITOSIS_ADMIN
fi

chown -R  $GITOSIS_USER *
cp $PUB_SSH_KEY ./keydir/$REPO_USER.pub
$CMD_GIT add  ./keydir/$REPO_USER.pub
output debug "Adding user '$REPO_USER' to writable list of gitosis-admin repo"
sed -ibak "s/^members.*$/& $REPO_USER/" gitosis.conf
$CMD_GIT commit -a -m "Added user '$REPO_USER' as writable to gitosis-admin repo"
run_cmd "cd $TEMP_DIR/$GITOSIS_ADMIN; $CMD_GIT push"

output debug "Finished setting up GITOSIS for initial use"

cd $PWD
